from .models import NaturalKeyModel, NaturalKeyModelManager, NaturalKeyQuerySet

try:
    from .serializers import NaturalKeySerializer, NaturalKeyModelSerializer
except ImportError:
    NaturalKeySerializer = None
    NaturalKeyModelSerializer = None


__all__ = [
    "NaturalKeyModel",
    "NaturalKeyModelManager",
    "NaturalKeySerializer",
    "NaturalKeyModelSerializer",
    "NaturalKeyQuerySet",
]



import os
from django.test.utils import setup_test_environment
import django
from django.core.management import call_command

os.environ["DJANGO_SETTINGS_MODULE"] = "tests.settings"
setup_test_environment()
django.setup()
call_command("makemigrations", interactive=False)
call_command("migrate", interactive=False)



try:
    from rest_framework import viewsets
    from rest_framework import serializers
except ImportError:
    pass
else:
    from .models import (
        NaturalKeyChild,
        ModelWithNaturalKey,
        ModelWithSingleUniqueField,
    )
    from natural_keys import NaturalKeySerializer, NaturalKeyModelSerializer

    class NaturalKeyChildViewSet(viewsets.ModelViewSet):
        queryset = NaturalKeyChild.objects.all()
        serializer_class = NaturalKeySerializer.for_model(NaturalKeyChild)

    class ModelWithNaturalKeyViewSet(viewsets.ModelViewSet):
        queryset = ModelWithNaturalKey.objects.all()
        serializer_class = NaturalKeyModelSerializer.for_model(
            ModelWithNaturalKey,
            include_fields="__all__",
        )

    class ModelWithSingleUniqueFieldViewSet(viewsets.ModelViewSet):
        queryset = ModelWithSingleUniqueField.objects.all()
        serializer_class = NaturalKeyModelSerializer.for_model(
            ModelWithSingleUniqueField,
            include_fields="__all__",
        )

    class LookupSerializer(NaturalKeyModelSerializer):
        id = serializers.ReadOnlyField(source="natural_key_slug")

        class Meta:
            model = NaturalKeyChild
            fields = "__all__"

    class NaturalKeyLookupViewSet(viewsets.ModelViewSet):
        queryset = NaturalKeyChild.objects.all()
        serializer_class = LookupSerializer
        lookup_field = "natural_key_slug"



try:
    from rest_framework import routers
except ImportError:
    urlpatterns = []
else:
    from .views import (
        NaturalKeyChildViewSet,
        ModelWithNaturalKeyViewSet,
        ModelWithSingleUniqueFieldViewSet,
        NaturalKeyLookupViewSet,
    )

    router = routers.DefaultRouter()
    router.register(r"naturalkeychilds", NaturalKeyChildViewSet)
    router.register(r"modelwithnaturalkeys", ModelWithNaturalKeyViewSet)
    router.register(
        r"modelwithsingleuniquefield", ModelWithSingleUniqueFieldViewSet
    )
    router.register(
        r"naturalkeylookup", NaturalKeyLookupViewSet, basename="lookup"
    )

    urlpatterns = router.urls



from rest_framework import serializers
from rest_framework.utils import model_meta
from rest_framework.validators import UniqueValidator
from html_json_forms.serializers import JSONFormModelSerializer
from .models import NaturalKeyModel
from collections import OrderedDict


class NaturalKeyValidator(serializers.UniqueTogetherValidator):
    def set_context(self, serializer):
        if getattr(self, "requires_context", None):
            # DRF 3.11+
            pass
        else:
            # DRF 3.10 and older
            self.serializer = serializer
            super(NaturalKeyValidator, self).set_context(serializer)

    def filter_queryset(self, attrs, queryset, serializer=None):
        if not serializer:
            # DRF 3.10 and older
            serializer = self.serializer

        nested_fields = {
            name: serializer.fields[name]
            for name in self.fields
            if isinstance(serializer.fields[name], NaturalKeySerializer)
        }

        attrs = attrs.copy()
        for field in attrs:
            if field in nested_fields:
                assert isinstance(attrs[field], dict)
                cls = nested_fields[field].Meta.model
                result = cls._default_manager.filter(**attrs[field])
                if result.count() == 0:
                    # No existing nested object for these values
                    return queryset.none()
                else:
                    # Existing nested object, use it to validate
                    attrs[field] = result[0].pk

        if getattr(self, "requires_context", None):
            # DRF 3.11+
            return super(NaturalKeyValidator, self).filter_queryset(
                attrs, queryset, serializer
            )
        else:
            # DRF 3.10 and older
            return super(NaturalKeyValidator, self).filter_queryset(
                attrs, queryset
            )


class NaturalKeySerializer(JSONFormModelSerializer):
    """
    Self-nesting Serializer for NaturalKeyModels
    """

    def build_standard_field(self, field_name, model_field):
        field_class, field_kwargs = super(
            NaturalKeySerializer, self
        ).build_standard_field(field_name, model_field)

        if "validators" in field_kwargs:
            field_kwargs["validators"] = [
                validator
                for validator in field_kwargs.get("validators", [])
                if not isinstance(validator, UniqueValidator)
            ]
        return field_class, field_kwargs

    def build_nested_field(self, field_name, relation_info, nested_depth):
        field_class = NaturalKeySerializer.for_model(
            relation_info.related_model,
            validate_key=False,
        )
        field_kwargs = {}
        return field_class, field_kwargs

    def create(self, validated_data):
        model_class = self.Meta.model
        natural_key_fields = model_class.get_natural_key_fields()
        natural_key = []
        for field in natural_key_fields:
            val = validated_data
            for key in field.split("__"):
                val = val[key]
            natural_key.append(val)
        return model_class.objects.find(*natural_key)

    def update(self, instance, validated_data):
        raise NotImplementedError(
            "Updating an existing natural key is not supported."
        )

    @classmethod
    def for_model(cls, model_class, validate_key=True, include_fields=None):
        unique_together = model_class.get_natural_key_def()
        if include_fields and list(include_fields) != list(unique_together):
            raise NotImplementedError(
                "NaturalKeySerializer for '%s' has unique_together = [%s], "
                "but provided include_fields = [%s]"
                % (
                    model_class._meta.model_name,
                    ", ".join(unique_together),
                    ", ".join(include_fields),
                )
            )

        class Serializer(cls):
            class Meta(cls.Meta):
                model = model_class
                fields = unique_together
                if validate_key:
                    validators = [
                        NaturalKeyValidator(
                            queryset=model_class._default_manager,
                            fields=unique_together,
                        )
                    ]
                else:
                    validators = []

        return Serializer

    @classmethod
    def for_depth(cls, model_class):
        return cls

    class Meta:
        depth = 1


class NaturalKeyModelSerializer(JSONFormModelSerializer):
    """
    Serializer for models with one or more foreign keys to a NaturalKeyModel
    """

    def is_natural_key_model(self, related_model):
        return issubclass(related_model, NaturalKeyModel)

    def build_nested_field(self, field_name, relation_info, nested_depth):
        if self.is_natural_key_model(relation_info.related_model):
            field_class = NaturalKeySerializer.for_model(
                relation_info.related_model,
                validate_key=False,
            )
            field_kwargs = {}
            if relation_info.model_field.null:
                field_kwargs["required"] = False
            return field_class, field_kwargs

        return super(NaturalKeyModelSerializer, self).build_nested_field(
            field_name, relation_info, nested_depth
        )

    def build_relational_field(self, field_name, relation_info):
        field_class, field_kwargs = super(
            NaturalKeyModelSerializer, self
        ).build_relational_field(field_name, relation_info)
        if self.is_natural_key_model(relation_info.related_model):
            field_kwargs.pop("queryset")
            field_kwargs["read_only"] = True
        return field_class, field_kwargs

    def get_fields(self):
        fields = super(NaturalKeyModelSerializer, self).get_fields()
        fields.update(self.build_natural_key_fields())
        return fields

    def build_natural_key_fields(self):
        info = model_meta.get_field_info(self.Meta.model)
        fields = OrderedDict()
        for field, relation_info in info.relations.items():
            if relation_info.reverse or relation_info.to_many:
                continue
            if not self.is_natural_key_model(relation_info.related_model):
                continue
            field_class, field_kwargs = self.build_nested_field(
                field, relation_info, 1
            )
            fields[field] = field_class(**field_kwargs)
        return fields

    def create(self, validated_data):
        self.convert_natural_keys(validated_data)
        return super(NaturalKeyModelSerializer, self).create(validated_data)

    def update(self, instance, validated_data):
        self.convert_natural_keys(validated_data)
        return super(NaturalKeyModelSerializer, self).update(
            instance, validated_data
        )

    def convert_natural_keys(self, validated_data):
        fields = self.get_fields()
        for name, field in fields.items():
            if name not in validated_data:
                continue
            if isinstance(field, NaturalKeySerializer):
                validated_data[name] = fields[name].create(
                    validated_data[name]
                )

    @classmethod
    def for_model(cls, model_class, include_fields=None):
        # c.f. wq.db.rest.serializers.ModelSerializer
        class Serializer(cls):
            class Meta(cls.Meta):
                model = model_class
                if include_fields:
                    fields = include_fields

        return Serializer

    class Meta:
        pass



from django.db import models
from natural_keys import NaturalKeyModel


class NaturalKeyParent(NaturalKeyModel):
    code = models.CharField(max_length=10)
    group = models.CharField(max_length=10)

    class Meta:
        unique_together = ["code", "group"]


class NaturalKeyChild(NaturalKeyModel):
    parent = models.ForeignKey(
        NaturalKeyParent,
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    mode = models.CharField(max_length=10)

    class Meta:
        unique_together = ["parent", "mode"]


class ModelWithNaturalKey(models.Model):
    key = models.ForeignKey(NaturalKeyChild, on_delete=models.CASCADE)
    value = models.CharField(max_length=10)


class ModelWithSingleUniqueField(NaturalKeyModel):
    code = models.CharField(max_length=10, unique=True)


class ModelWithExtraField(NaturalKeyModel):
    code = models.CharField(max_length=10)
    date = models.DateField(max_length=10)
    extra = models.TextField()

    class Meta:
        unique_together = ["code", "date"]


class ModelWithConstraint(NaturalKeyModel):
    code = models.CharField(max_length=10)
    date = models.DateField(max_length=10)

    def __str__(self):
        return f"{self.code} {self.date}"

    class Meta:
        constraints = [
            models.UniqueConstraint(
                name="natural key",
                fields=["code", "date"],
            ),
        ]



import unittest

try:
    from rest_framework.test import APITestCase
except ImportError:
    raise unittest.SkipTest("Skipping DRF tests as DRF is not installed.")

from rest_framework import status
from tests.test_app.models import (
    NaturalKeyChild,
    ModelWithNaturalKey,
    ModelWithSingleUniqueField,
)
from natural_keys import NaturalKeySerializer

# Tests for natural key DRF integration


class NaturalKeyRestTestCase(APITestCase):
    def test_naturalkey_rest_serializer(self):
        # Serializer should include validator
        serializer = NaturalKeySerializer.for_model(NaturalKeyChild)()
        expect = """
             Serializer():
                 parent = Serializer():
                     code = CharField(max_length=10)
                     group = CharField(max_length=10)
                 mode = CharField(max_length=10)
                 class Meta:
                     validators = [<NaturalKeyValidator(queryset=NaturalKeyChild.objects.all(), fields=('parent', 'mode'))>]""".replace(
            "             ", ""
        )[
            1:
        ]  # noqa
        self.assertEqual(expect, str(serializer))

        fields = serializer.get_fields()
        self.assertTrue(fields["parent"].required)
        self.assertTrue(fields["mode"].required)
        self.assertTrue(fields["parent"].get_fields()["code"].required)

    def test_naturalkey_rest_singleunique(self):
        # Serializer should only have single top-level validator
        serializer = NaturalKeySerializer.for_model(
            ModelWithSingleUniqueField
        )()
        expect = """
             Serializer():
                 code = CharField(max_length=10, validators=[])
                 class Meta:
                     validators = [<NaturalKeyValidator(queryset=ModelWithSingleUniqueField.objects.all(), fields=('code',))>]""".replace(
            "             ", ""
        )[
            1:
        ]  # noqa
        self.assertEqual(expect, str(serializer))

        fields = serializer.get_fields()
        self.assertTrue(fields["code"].required)

    def test_naturalkey_rest_post(self):
        # Posting a compound natural key should work
        form = {
            "mode": "mode3a",
            "parent[code]": "code3",
            "parent[group]": "group3",
        }
        response = self.client.post("/naturalkeychilds.json", form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )
        self.assertEqual(response.data["mode"], "mode3a")
        self.assertEqual(response.data["parent"]["code"], "code3")
        self.assertEqual(response.data["parent"]["group"], "group3")

        # Posting a simple natural key should work
        form = {
            "code": "code9",
        }
        response = self.client.post("/modelwithsingleuniquefield.json", form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )
        self.assertEqual(response.data["code"], "code9")

        # Posting same nested natural key should reuse nested object
        form = {
            "mode": "mode3b",
            "parent[code]": "code3",
            "parent[group]": "group3",
        }
        response = self.client.post("/naturalkeychilds.json", form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )
        self.assertEqual(
            NaturalKeyChild.objects.get(mode="mode3a").parent.pk,
            NaturalKeyChild.objects.get(mode="mode3b").parent.pk,
        )

    def test_naturalkey_rest_duplicate(self):
        # Posting identical compound natural key should fail
        form = {
            "mode": "mode3c",
            "parent[code]": "code3",
            "parent[group]": "group3",
        }
        response = self.client.post("/naturalkeychilds.json", form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )
        form = {
            "mode": "mode3c",
            "parent[code]": "code3",
            "parent[group]": "group3",
        }
        response = self.client.post("/naturalkeychilds.json", form)
        self.assertEqual(
            response.status_code, status.HTTP_400_BAD_REQUEST, response.data
        )
        self.assertEqual(
            response.data,
            {
                "non_field_errors": [
                    "The fields parent, mode must make a unique set."
                ]
            },
        )

        # Posting identical simple natural key should fail
        form = {
            "code": "code8",
        }
        response = self.client.post("/modelwithsingleuniquefield.json", form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )
        form = {
            "code": "code8",
        }
        response = self.client.post("/modelwithsingleuniquefield.json", form)
        self.assertEqual(
            response.status_code, status.HTTP_400_BAD_REQUEST, response.data
        )
        self.assertEqual(
            response.data,
            {
                "code": [
                    "model with single unique field "
                    "with this code already exists."
                ]
            },
        )

    def test_naturalkey_rest_nested_post(self):
        # Posting a regular model with a ref to natural key
        form = {
            "key[mode]": "mode4",
            "key[parent][code]": "code4",
            "key[parent][group]": "group4",
            "value": 5,
        }
        response = self.client.post("/modelwithnaturalkeys.json", form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )
        self.assertEqual(response.data["key"]["mode"], "mode4")
        self.assertEqual(response.data["key"]["parent"]["code"], "code4")
        self.assertEqual(response.data["key"]["parent"]["group"], "group4")

    def test_naturalkey_rest_nested_put(self):
        # Updating a regular model with a ref to natural key
        instance = ModelWithNaturalKey.objects.create(
            key=NaturalKeyChild.objects.find("code5", "group5", "mode5"),
            value=7,
        )
        self.assertEqual(instance.key.parent.code, "code5")

        # Updating with same natural key should reuse it
        form = {
            "key[mode]": "mode5",
            "key[parent][code]": "code5",
            "key[parent][group]": "group5",
            "value": 8,
        }
        self.assertEqual(NaturalKeyChild.objects.count(), 1)

        # Updating with new natural key should create it
        response = self.client.put(
            "/modelwithnaturalkeys/%s.json" % instance.pk, form
        )
        form = {
            "key[mode]": "mode6",
            "key[parent][code]": "code6",
            "key[parent][group]": "group6",
            "value": 9,
        }
        response = self.client.put(
            "/modelwithnaturalkeys/%s.json" % instance.pk, form
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.data
        )
        self.assertEqual(response.data["key"]["mode"], "mode6")
        self.assertEqual(response.data["key"]["parent"]["code"], "code6")
        self.assertEqual(response.data["key"]["parent"]["group"], "group6")
        self.assertEqual(NaturalKeyChild.objects.count(), 2)

    def test_naturalkey_lookup(self):
        # Support natural_key_slug as lookup_field setting
        NaturalKeyChild.objects.find("code7", "group7", "mode7")
        response = self.client.get(
            "/naturalkeylookup/code7-group7-mode7.json",
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.data
        )
        self.assertEqual(response.data["id"], "code7-group7-mode7")

    def test_naturalkey_lookup_slug(self):
        # Support separator in slug (but only for last part of key)
        NaturalKeyChild.objects.find("code7", "group7", "mode7-alt")
        response = self.client.get(
            "/naturalkeylookup/code7-group7-mode7-alt.json",
        )
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.data
        )
        self.assertEqual(response.data["id"], "code7-group7-mode7-alt")

    def test_invalid_slug_404(self):
        response = self.client.get(
            "/naturalkeylookup/not-valid.json",
        )
        self.assertEqual(
            status.HTTP_404_NOT_FOUND,
            response.status_code,
        )



SECRET_KEY = "1234"
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}
ROOT_URLCONF = "tests.urls"
REST_FRAMEWORK = {
    "UNAUTHENTICATED_USER": None,
}

INSTALLED_APPS = [
    "tests.test_app",
]



from django.test import TestCase
from tests.test_app.models import (
    NaturalKeyParent,
    NaturalKeyChild,
    ModelWithSingleUniqueField,
    ModelWithExtraField,
    ModelWithConstraint,
)
from django.db.utils import IntegrityError

# Tests for natural key models


class NaturalKeyTestCase(TestCase):
    def test_naturalkey_fields(self):
        # Model APIs
        self.assertEqual(
            NaturalKeyParent.get_natural_key_fields(),
            ("code", "group"),
        )
        self.assertEqual(
            NaturalKeyParent(code="code0", group="group0").natural_key(),
            ("code0", "group0"),
        )
        self.assertEqual(
            NaturalKeyChild.get_natural_key_fields(),
            ("parent__code", "parent__group", "mode"),
        )
        self.assertEqual(
            ModelWithSingleUniqueField.get_natural_key_fields(),
            ("code",),
        )

    def test_naturalkey_create(self):
        # Manager create
        p1 = NaturalKeyParent.objects.create_by_natural_key("code1", "group1")
        self.assertEqual(p1.code, "code1")
        self.assertEqual(p1.group, "group1")

        # get_or_create with same key retrieve existing item
        p2, is_new = NaturalKeyParent.objects.get_or_create_by_natural_key(
            "code1", "group1"
        )
        self.assertFalse(is_new)
        self.assertEqual(p1.pk, p2.pk)

        # Shortcut version
        p3 = NaturalKeyParent.objects.find("code1", "group1")
        self.assertEqual(p1.pk, p3.pk)

        p4 = ModelWithSingleUniqueField.objects.create_by_natural_key("code4")
        self.assertEqual(p4.code, "code4")

    def test_naturalkey_nested_create(self):
        # Manager create, with nested natural key
        c1 = NaturalKeyChild.objects.create_by_natural_key(
            "code2", "group2", "mode1"
        )
        self.assertEqual(c1.parent.code, "code2")
        self.assertEqual(c1.parent.group, "group2")
        self.assertEqual(c1.mode, "mode1")

        # create with same nested key should not create a new parent
        c2 = NaturalKeyChild.objects.create_by_natural_key(
            "code2", "group2", "mode2"
        )
        self.assertEqual(c1.parent.pk, c2.parent.pk)
        self.assertEqual(c2.mode, "mode2")

    def test_naturalkey_duplicate(self):
        # Manager create, with duplicate
        NaturalKeyParent.objects.create_by_natural_key("code1", "group1")
        # create with same key should fail
        with self.assertRaises(IntegrityError):
            NaturalKeyParent.objects.create_by_natural_key("code1", "group1")

    def test_filter_with_Q(self):
        from django.db.models import Q

        query = Q(code="bizarre")
        self.assertEqual(
            ModelWithSingleUniqueField.objects.filter(query).count(), 0
        )

    def test_find_with_defaults(self):
        obj = ModelWithExtraField.objects.find(
            "extra1",
            "2019-07-26",
            defaults={"extra": "Test 123"},
        )
        self.assertEqual(obj.extra, "Test 123")

    def test_find_with_kwargs(self):
        with self.assertRaises(TypeError) as e:
            ModelWithExtraField.objects.find(
                "extra1",
                date="2019-07-26",
            )
        msg = str(e.exception).replace("NaturalKeyModelManager.", "")
        self.assertEqual(
            msg, "find() got an unexpected keyword argument 'date'"
        )

    def test_find_with_constraint(self):
        obj = ModelWithConstraint.objects.find(
            "constraint1",
            "2021-08-23",
        )
        self.assertEqual(str(obj), "constraint1 2021-08-23")

    def test_null_foreign_key(self):
        obj = NaturalKeyChild.objects.create(mode="mode0")
        self.assertEqual(
            obj.natural_key(),
            (
                None,
                None,
                "mode0",
            ),
        )
        self.assertEqual(
            NaturalKeyChild.objects.get_by_natural_key(
                None,
                None,
                "mode0",
            ),
            obj,
        )



# Generated by Django 4.2.2 on 2023-06-15 12:37

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="ModelWithConstraint",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("code", models.CharField(max_length=10)),
                ("date", models.DateField(max_length=10)),
            ],
        ),
        migrations.CreateModel(
            name="ModelWithSingleUniqueField",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("code", models.CharField(max_length=10, unique=True)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="NaturalKeyParent",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("code", models.CharField(max_length=10)),
                ("group", models.CharField(max_length=10)),
            ],
            options={
                "unique_together": {("code", "group")},
            },
        ),
        migrations.CreateModel(
            name="NaturalKeyChild",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("mode", models.CharField(max_length=10)),
                (
                    "parent",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="test_app.naturalkeyparent",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ModelWithNaturalKey",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("value", models.CharField(max_length=10)),
                (
                    "key",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="test_app.naturalkeychild",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ModelWithExtraField",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("code", models.CharField(max_length=10)),
                ("date", models.DateField(max_length=10)),
                ("extra", models.TextField()),
            ],
            options={
                "unique_together": {("code", "date")},
            },
        ),
        migrations.AddConstraint(
            model_name="modelwithconstraint",
            constraint=models.UniqueConstraint(
                fields=("code", "date"), name="natural key"
            ),
        ),
        migrations.AlterUniqueTogether(
            name="naturalkeychild",
            unique_together={("parent", "mode")},
        ),
    ]



from django.db import models
from functools import reduce


class NaturalKeyQuerySet(models.QuerySet):
    def filter(self, *args, **kwargs):
        natural_key_slug = kwargs.pop("natural_key_slug", None)
        if natural_key_slug and type(natural_key_slug) is str:
            slugs = natural_key_slug.split(self.model.natural_key_separator)
            fields = self.model.get_natural_key_fields()
            if len(slugs) > len(fields):
                slugs[len(fields) - 1 :] = [
                    self.model.natural_key_separator.join(
                        slugs[len(fields) - 1 :]
                    )
                ]
            elif len(slugs) < len(fields):
                return self.none()
            kwargs.update(self.natural_key_kwargs(*slugs))

        return super(NaturalKeyQuerySet, self).filter(*args, **kwargs)

    def natural_key_kwargs(self, *args):
        natural_key = self.model.get_natural_key_fields()
        if len(args) != len(natural_key):
            raise TypeError(
                "Wrong number of values, expected %s" % len(natural_key)
            )
        return dict(zip(natural_key, args))


class NaturalKeyModelManager(models.Manager):
    """
    Manager for use with subclasses of NaturalKeyModel.
    """

    def get_queryset(self):
        return NaturalKeyQuerySet(self.model, using=self._db)

    def get_by_natural_key(self, *args):
        """
        Return the object corresponding to the provided natural key.

        (This is a generic implementation of the standard Django function)
        """

        kwargs = self.natural_key_kwargs(*args)

        # Since kwargs already has __ lookups in it, we could just do this:
        # return self.get(**kwargs)

        # But, we should call each related model's get_by_natural_key in case
        # it's been overridden
        for name, rel_to in self.model.get_natural_key_info():
            if not rel_to:
                continue

            # Extract natural key for related object
            nested_key = extract_nested_key(kwargs, rel_to, name)
            if nested_key:
                # Update kwargs with related object
                try:
                    kwargs[name] = rel_to.objects.get_by_natural_key(
                        *nested_key
                    )
                except rel_to.DoesNotExist:
                    # If related object doesn't exist, assume this one doesn't
                    raise self.model.DoesNotExist()
            else:
                kwargs[name] = None

        return self.get(**kwargs)

    def create_by_natural_key(self, *args, defaults=None):
        """
        Create a new object from the provided natural key values.  If the
        natural key contains related objects, recursively get or create them by
        their natural keys.
        """

        kwargs = self.natural_key_kwargs(*args)
        for name, rel_to in self.model.get_natural_key_info():
            if not rel_to:
                continue
            nested_key = extract_nested_key(kwargs, rel_to, name)
            # Automatically create any related objects as needed
            if nested_key:
                (
                    kwargs[name],
                    is_new,
                ) = rel_to.objects.get_or_create_by_natural_key(*nested_key)
            else:
                kwargs[name] = None
        if defaults:
            attrs = defaults
            attrs.update(kwargs)
        else:
            attrs = kwargs
        return self.create(**attrs)

    def get_or_create_by_natural_key(self, *args, defaults=None):
        """
        get_or_create + get_by_natural_key
        """
        try:
            return (
                self.get_by_natural_key(*args),
                False,
            )
        except self.model.DoesNotExist:
            return (
                self.create_by_natural_key(
                    *args,
                    defaults=defaults,
                ),
                True,
            )

    # Shortcut for common use case
    def find(self, *args, defaults=None):
        """
        Shortcut for get_or_create_by_natural_key that discards the "created"
        boolean.
        """
        obj, is_new = self.get_or_create_by_natural_key(
            *args,
            defaults=defaults,
        )
        return obj

    def natural_key_kwargs(self, *args):
        """
        Convert args into kwargs by merging with model's natural key fieldnames
        """
        return self.get_queryset().natural_key_kwargs(*args)

    def resolve_keys(self, keys, auto_create=False):
        """
        Resolve the list of given keys into objects, if possible.
        Returns a mapping and a success indicator.
        """
        resolved = {}
        success = True
        for key in keys:
            if auto_create:
                resolved[key] = self.find(*key)
            else:
                try:
                    resolved[key] = self.get_by_natural_key(*key)
                except self.model.DoesNotExist:
                    success = False
                    resolved[key] = None
        return resolved, success


class NaturalKeyModel(models.Model):
    """
    Abstract class with a generic implementation of natural_key.
    """

    objects = NaturalKeyModelManager()

    @classmethod
    def get_natural_key_info(cls):
        """
        Derive natural key from first unique_together definition, noting which
        fields are related objects vs. regular fields.
        """
        fields = cls.get_natural_key_def()
        info = []
        for name in fields:
            field = cls._meta.get_field(name)
            rel_to = None
            if hasattr(field, "rel"):
                rel_to = field.rel.to if field.rel else None
            elif hasattr(field, "remote_field"):
                if field.remote_field:
                    rel_to = field.remote_field.model
                else:
                    rel_to = None
            info.append((name, rel_to))
        return info

    @classmethod
    def get_natural_key_def(cls):
        if hasattr(cls, "_natural_key"):
            return cls._natural_key

        for constraint in cls._meta.constraints:
            if isinstance(constraint, models.UniqueConstraint):
                return constraint.fields

        if cls._meta.unique_together:
            return cls._meta.unique_together[0]

        unique = [
            f
            for f in cls._meta.fields
            if f.unique
            and f.__class__.__name__ not in ["AutoField", "BigAutoField"]
        ]
        if unique:
            return (unique[0].name,)

        raise Exception("Add a UniqueConstraint to use natural-keys")

    @classmethod
    def get_natural_key_fields(cls):
        """
        Determine actual natural key field list, incorporating the natural keys
        of related objects as needed.
        """
        natural_key = []
        for name, rel_to in cls.get_natural_key_info():
            if not rel_to:
                natural_key.append(name)
            else:
                nested_key = rel_to.get_natural_key_fields()
                natural_key.extend(
                    [name + "__" + nname for nname in nested_key]
                )
        return tuple(natural_key)

    def natural_key(self):
        """
        Return the natural key for this object.

        (This is a generic implementation of the standard Django function)
        """
        return tuple(
            self.get_natural_key_value(field)
            for field in self.get_natural_key_fields()
        )

    def get_natural_key_value(self, field):
        # Recursively extract properties from related objects if needed
        obj = self
        for part in field.split("__"):
            obj = getattr(obj, part)
            if obj is None:
                return None
        return obj

    natural_key_separator = "-"

    @property
    def natural_key_slug(self):
        return self.natural_key_separator.join(
            str(slug) for slug in self.natural_key()
        )

    class Meta:
        abstract = True


def extract_nested_key(key, cls, prefix=""):
    nested_key = cls.get_natural_key_fields()
    local_fields = {field.name: field for field in cls._meta.local_fields}
    values = []
    has_val = False
    if prefix:
        prefix += "__"
    for nname in nested_key:
        val = key.pop(prefix + nname, None)
        if val is None and nname in local_fields:
            if type(local_fields[nname]).__name__ == "DateTimeField":
                date = key.pop(nname + "_date", None)
                time = key.pop(nname + "_time", None)
                if date and time:
                    val = "%s %s" % (date, time)

        if val is not None:
            has_val = True
        values.append(val)
    if has_val:
        return values
    else:
        return None
